# Generated from /home/frances/projects/code/tarski/utils/parsers/grammars/fstrips.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .parser import fstripsParser
else:
    from parser import fstripsParser

# This class defines a complete generic visitor for a parse tree produced by fstripsParser.

class fstripsVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by fstripsParser#pddlDoc.
    def visitPddlDoc(self, ctx:fstripsParser.PddlDocContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#domain.
    def visitDomain(self, ctx:fstripsParser.DomainContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#domainName.
    def visitDomainName(self, ctx:fstripsParser.DomainNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#requireDef.
    def visitRequireDef(self, ctx:fstripsParser.RequireDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#declaration_of_types.
    def visitDeclaration_of_types(self, ctx:fstripsParser.Declaration_of_typesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#Integer.
    def visitInteger(self, ctx:fstripsParser.IntegerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#Float.
    def visitFloat(self, ctx:fstripsParser.FloatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#Number.
    def visitNumber(self, ctx:fstripsParser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#NumericBuiltin.
    def visitNumericBuiltin(self, ctx:fstripsParser.NumericBuiltinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ObjectBuiltin.
    def visitObjectBuiltin(self, ctx:fstripsParser.ObjectBuiltinContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#SimpleNameList.
    def visitSimpleNameList(self, ctx:fstripsParser.SimpleNameListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ComplexNameList.
    def visitComplexNameList(self, ctx:fstripsParser.ComplexNameListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#name_list_with_type.
    def visitName_list_with_type(self, ctx:fstripsParser.Name_list_with_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#UntypedTypenameList.
    def visitUntypedTypenameList(self, ctx:fstripsParser.UntypedTypenameListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TypedTypenameList.
    def visitTypedTypenameList(self, ctx:fstripsParser.TypedTypenameListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#UntypedVariableList.
    def visitUntypedVariableList(self, ctx:fstripsParser.UntypedVariableListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TypedVariableList.
    def visitTypedVariableList(self, ctx:fstripsParser.TypedVariableListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#variable_list_with_type.
    def visitVariable_list_with_type(self, ctx:fstripsParser.Variable_list_with_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#EitherTypename.
    def visitEitherTypename(self, ctx:fstripsParser.EitherTypenameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#PrimitiveTypename.
    def visitPrimitiveTypename(self, ctx:fstripsParser.PrimitiveTypenameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#primitive_type.
    def visitPrimitive_type(self, ctx:fstripsParser.Primitive_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#function_definition_block.
    def visitFunction_definition_block(self, ctx:fstripsParser.Function_definition_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#single_function_definition.
    def visitSingle_function_definition(self, ctx:fstripsParser.Single_function_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#typed_function_definition.
    def visitTyped_function_definition(self, ctx:fstripsParser.Typed_function_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#untyped_function_definition.
    def visitUntyped_function_definition(self, ctx:fstripsParser.Untyped_function_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#logical_symbol_name.
    def visitLogical_symbol_name(self, ctx:fstripsParser.Logical_symbol_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#constant_declaration.
    def visitConstant_declaration(self, ctx:fstripsParser.Constant_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#predicate_definition_block.
    def visitPredicate_definition_block(self, ctx:fstripsParser.Predicate_definition_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#single_predicate_definition.
    def visitSingle_predicate_definition(self, ctx:fstripsParser.Single_predicate_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#predicate.
    def visitPredicate(self, ctx:fstripsParser.PredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#function_name.
    def visitFunction_name(self, ctx:fstripsParser.Function_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#structureDef.
    def visitStructureDef(self, ctx:fstripsParser.StructureDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#actionDef.
    def visitActionDef(self, ctx:fstripsParser.ActionDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#constraintDef.
    def visitConstraintDef(self, ctx:fstripsParser.ConstraintDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#eventDef.
    def visitEventDef(self, ctx:fstripsParser.EventDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#actionName.
    def visitActionName(self, ctx:fstripsParser.ActionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#constraintSymbol.
    def visitConstraintSymbol(self, ctx:fstripsParser.ConstraintSymbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#eventSymbol.
    def visitEventSymbol(self, ctx:fstripsParser.EventSymbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#actionDefBody.
    def visitActionDefBody(self, ctx:fstripsParser.ActionDefBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TrivialPrecondition.
    def visitTrivialPrecondition(self, ctx:fstripsParser.TrivialPreconditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#RegularPrecondition.
    def visitRegularPrecondition(self, ctx:fstripsParser.RegularPreconditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TermGoalDesc.
    def visitTermGoalDesc(self, ctx:fstripsParser.TermGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AndGoalDesc.
    def visitAndGoalDesc(self, ctx:fstripsParser.AndGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#OrGoalDesc.
    def visitOrGoalDesc(self, ctx:fstripsParser.OrGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#NotGoalDesc.
    def visitNotGoalDesc(self, ctx:fstripsParser.NotGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ImplyGoalDesc.
    def visitImplyGoalDesc(self, ctx:fstripsParser.ImplyGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ExistentialGoalDesc.
    def visitExistentialGoalDesc(self, ctx:fstripsParser.ExistentialGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#UniversalGoalDesc.
    def visitUniversalGoalDesc(self, ctx:fstripsParser.UniversalGoalDescContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#BuiltinBinaryAtom.
    def visitBuiltinBinaryAtom(self, ctx:fstripsParser.BuiltinBinaryAtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#atomicTermFormula.
    def visitAtomicTermFormula(self, ctx:fstripsParser.AtomicTermFormulaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TermObject.
    def visitTermObject(self, ctx:fstripsParser.TermObjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TermNumber.
    def visitTermNumber(self, ctx:fstripsParser.TermNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TermVariable.
    def visitTermVariable(self, ctx:fstripsParser.TermVariableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TermTimeStep.
    def visitTermTimeStep(self, ctx:fstripsParser.TermTimeStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TermFunction.
    def visitTermFunction(self, ctx:fstripsParser.TermFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#GenericFunctionTerm.
    def visitGenericFunctionTerm(self, ctx:fstripsParser.GenericFunctionTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#BinaryArithmeticFunctionTerm.
    def visitBinaryArithmeticFunctionTerm(self, ctx:fstripsParser.BinaryArithmeticFunctionTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#UnaryArithmeticFunctionTerm.
    def visitUnaryArithmeticFunctionTerm(self, ctx:fstripsParser.UnaryArithmeticFunctionTermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#derivedDef.
    def visitDerivedDef(self, ctx:fstripsParser.DerivedDefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ConjunctiveEffectFormula.
    def visitConjunctiveEffectFormula(self, ctx:fstripsParser.ConjunctiveEffectFormulaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#SingleEffect.
    def visitSingleEffect(self, ctx:fstripsParser.SingleEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#UniversallyQuantifiedEffect.
    def visitUniversallyQuantifiedEffect(self, ctx:fstripsParser.UniversallyQuantifiedEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#SingleConditionalEffect.
    def visitSingleConditionalEffect(self, ctx:fstripsParser.SingleConditionalEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#MultipleConditionalEffect.
    def visitMultipleConditionalEffect(self, ctx:fstripsParser.MultipleConditionalEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AtomicEffect.
    def visitAtomicEffect(self, ctx:fstripsParser.AtomicEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AssignConstant.
    def visitAssignConstant(self, ctx:fstripsParser.AssignConstantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AssignEffect.
    def visitAssignEffect(self, ctx:fstripsParser.AssignEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#DeleteAtomEffect.
    def visitDeleteAtomEffect(self, ctx:fstripsParser.DeleteAtomEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AddAtomEffect.
    def visitAddAtomEffect(self, ctx:fstripsParser.AddAtomEffectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#builtin_binary_function.
    def visitBuiltin_binary_function(self, ctx:fstripsParser.Builtin_binary_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#builtin_unary_function.
    def visitBuiltin_unary_function(self, ctx:fstripsParser.Builtin_unary_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#builtin_binary_predicate.
    def visitBuiltin_binary_predicate(self, ctx:fstripsParser.Builtin_binary_predicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#assignOp.
    def visitAssignOp(self, ctx:fstripsParser.AssignOpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#problem.
    def visitProblem(self, ctx:fstripsParser.ProblemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#problemMeta.
    def visitProblemMeta(self, ctx:fstripsParser.ProblemMetaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#problemDecl.
    def visitProblemDecl(self, ctx:fstripsParser.ProblemDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#problemDomain.
    def visitProblemDomain(self, ctx:fstripsParser.ProblemDomainContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#object_declaration.
    def visitObject_declaration(self, ctx:fstripsParser.Object_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#boundsDecl.
    def visitBoundsDecl(self, ctx:fstripsParser.BoundsDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#typeBoundsDefinition.
    def visitTypeBoundsDefinition(self, ctx:fstripsParser.TypeBoundsDefinitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#init.
    def visitInit(self, ctx:fstripsParser.InitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#InitPositiveLiteral.
    def visitInitPositiveLiteral(self, ctx:fstripsParser.InitPositiveLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#InitNegativeLiteral.
    def visitInitNegativeLiteral(self, ctx:fstripsParser.InitNegativeLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#InitFunctionAssignment.
    def visitInitFunctionAssignment(self, ctx:fstripsParser.InitFunctionAssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#flat_term.
    def visitFlat_term(self, ctx:fstripsParser.Flat_termContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#flat_atom.
    def visitFlat_atom(self, ctx:fstripsParser.Flat_atomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#symbolic_constant.
    def visitSymbolic_constant(self, ctx:fstripsParser.Symbolic_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#numeric_constant.
    def visitNumeric_constant(self, ctx:fstripsParser.Numeric_constantContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#goal.
    def visitGoal(self, ctx:fstripsParser.GoalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#probConstraints.
    def visitProbConstraints(self, ctx:fstripsParser.ProbConstraintsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ConjunctionOfConstraints.
    def visitConjunctionOfConstraints(self, ctx:fstripsParser.ConjunctionOfConstraintsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#UniversallyQuantifiedConstraint.
    def visitUniversallyQuantifiedConstraint(self, ctx:fstripsParser.UniversallyQuantifiedConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#PreferenceConstraint.
    def visitPreferenceConstraint(self, ctx:fstripsParser.PreferenceConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#PlainConstraintList.
    def visitPlainConstraintList(self, ctx:fstripsParser.PlainConstraintListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ProblemMetric.
    def visitProblemMetric(self, ctx:fstripsParser.ProblemMetricContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#optimization.
    def visitOptimization(self, ctx:fstripsParser.OptimizationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#FunctionalExprMetric.
    def visitFunctionalExprMetric(self, ctx:fstripsParser.FunctionalExprMetricContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#CompositeMetric.
    def visitCompositeMetric(self, ctx:fstripsParser.CompositeMetricContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#TotalTimeMetric.
    def visitTotalTimeMetric(self, ctx:fstripsParser.TotalTimeMetricContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#IsViolatedMetric.
    def visitIsViolatedMetric(self, ctx:fstripsParser.IsViolatedMetricContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#terminalCost.
    def visitTerminalCost(self, ctx:fstripsParser.TerminalCostContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#stageCost.
    def visitStageCost(self, ctx:fstripsParser.StageCostContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ConjunctiveConstraint.
    def visitConjunctiveConstraint(self, ctx:fstripsParser.ConjunctiveConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#ForallConstraint.
    def visitForallConstraint(self, ctx:fstripsParser.ForallConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AtEndConstraint.
    def visitAtEndConstraint(self, ctx:fstripsParser.AtEndConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AlwaysConstraint.
    def visitAlwaysConstraint(self, ctx:fstripsParser.AlwaysConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#SometimeConstraint.
    def visitSometimeConstraint(self, ctx:fstripsParser.SometimeConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#WithinConstraint.
    def visitWithinConstraint(self, ctx:fstripsParser.WithinConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AtMostOnceConstraint.
    def visitAtMostOnceConstraint(self, ctx:fstripsParser.AtMostOnceConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#SometimeAfterConstraint.
    def visitSometimeAfterConstraint(self, ctx:fstripsParser.SometimeAfterConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#SometimeBeforeConstraint.
    def visitSometimeBeforeConstraint(self, ctx:fstripsParser.SometimeBeforeConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AlwaysWithinConstraint.
    def visitAlwaysWithinConstraint(self, ctx:fstripsParser.AlwaysWithinConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#HoldDuringConstraint.
    def visitHoldDuringConstraint(self, ctx:fstripsParser.HoldDuringConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#HoldAfterConstraint.
    def visitHoldAfterConstraint(self, ctx:fstripsParser.HoldAfterConstraintContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by fstripsParser#AlternativeAlwaysConstraint.
    def visitAlternativeAlwaysConstraint(self, ctx:fstripsParser.AlternativeAlwaysConstraintContext):
        return self.visitChildren(ctx)



del fstripsParser


import operator
from .. import funcsym
from .. import errors as err
from ..syntax import ops, Connective, Atom, Formula, CompoundFormula, QuantifiedFormula, builtins, Variable, \
    Constant, CompoundTerm, Tautology, Contradiction, IfThenElse, AggregateCompoundTerm
from ..syntax.algebra import Matrix
from ..model import Model


# TODO We will need to extend this so that the interpretation depends on a certain, given sigma of values to
# TODO the free variables in the given termn / sentence, as is standard in textbook-FOL
def evaluate(element, m: Model, sigma=None):
    """ Evaluate the denotation of a given formula or term over a given model """
    sigma = sigma if sigma is not None else {}

    # Formulas
    if isinstance(element, Tautology):
        return True

    if isinstance(element, Contradiction):
        return False

    if isinstance(element, Atom):
        return evaluate_atom(element, m, sigma)

    if isinstance(element, CompoundFormula):
        return _compound_evaluators[element.connective](element, m, sigma)

    if isinstance(element, QuantifiedFormula):
        return evaluate_quantified(element, m, sigma)

    # Terms
    if isinstance(element, Variable):
        return sigma[element]  # TODO Finish this, ATM it will raise a runtime error

    if isinstance(element, (Constant, CompoundTerm, IfThenElse, Matrix, AggregateCompoundTerm)):
        return evaluate_term(element, m, sigma)

    raise RuntimeError("Unknown logical element type: {}, {}".format(element, type(element)))


_compound_evaluators = {
    Connective.Not: lambda f, m, s: not evaluate(f.subformulas[0], m, s),
    Connective.And: lambda f, m, s: all(evaluate(sub, m, s) for sub in f.subformulas),
    Connective.Or: lambda f, m, s: any(evaluate(sub, m, s) for sub in f.subformulas)
}


def evaluate_atom(atom: Atom, m: Model, sigma):
    if builtins.is_builtin_predicate(atom.predicate):
        return evaluate_builtin_predicate(atom, m, sigma)

    # Otherwise, the extension is given by the model
    point = tuple(evaluate(t, m, sigma) for t in atom.subterms)
    return m.holds(atom.predicate, point)


def evaluate_quantified(formula: Formula, m: Model, sigma):
    # pylint: disable=unused-argument
    raise NotImplementedError()


def evaluate_term(term, m: Model, sigma):
    if isinstance(term, IfThenElse):
        if evaluate(term.condition, m, sigma):  # condition is true
            term = term.subterms[0]
        else:
            term = term.subterms[1]

    if isinstance(term, Matrix):
        N, M = term.shape
        result = []
        for i in range(N):
            row = [evaluate_term(term.matrix[i, j], m, sigma) for j in range(M)]
            result.append(row)
        return Matrix(result, term.sort)

    if isinstance(term, CompoundTerm) and builtins.is_builtin_function(term.symbol):
        return evaluate_builtin_function(term, m, sigma)
    # MRJ: Coerce float and int Python literals into constants
    if isinstance(term, float):
        term = Constant(term, m.language.Real)
    if isinstance(term, int):
        term = Constant(term, m.language.Integer)
    assert isinstance(term, (Constant, CompoundTerm))  # TODO Add support for variables
    arguments = []

    # TODO This is technically not correct - a constant still depends on the actual model on which is evaluated.
    # TODO A constant is simply a nullary compound term. E.g. an integer variable is still a logical constant.
    # TODO "Constant" does NOT mean "fixed denotation", although we can assume for convenience that there is one
    # TODO constant for each element in the universe, and the other way round
    # TODO How does Z3 deal with this distinction?
    # TODO We'll need to fix this eventually
    if isinstance(term, Constant):
        return term

    if isinstance(term, CompoundTerm):
        # evaluate each of the arguments
        arguments = tuple(evaluate(a, m, sigma) for a in term.subterms)

    try:
        return m.value(term.symbol, arguments)
    except KeyError:
        raise err.UndefinedTerm(term)


def evaluate_builtin_predicate(atom, model, sigma):
    bip = builtins.BuiltinPredicateSymbol
    _evaluators = {
        bip.EQ: lambda f, m, s: evaluate(f.subterms[0], m, s).symbol == evaluate(f.subterms[1], m, s).symbol,
        bip.NE: lambda f, m, s: evaluate(f.subterms[0], m, s).symbol != evaluate(f.subterms[1], m, s).symbol,
        bip.LT: lambda f, m, s: evaluate(f.subterms[0], m, s).symbol < evaluate(f.subterms[1], m, s).symbol,
        bip.LE: lambda f, m, s: evaluate(f.subterms[0], m, s).symbol <= evaluate(f.subterms[1], m, s).symbol,
        bip.GT: lambda f, m, s: evaluate(f.subterms[0], m, s).symbol > evaluate(f.subterms[1], m, s).symbol,
        bip.GE: lambda f, m, s: evaluate(f.subterms[0], m, s).symbol >= evaluate(f.subterms[1], m, s).symbol,
    }

    return _evaluators[atom.predicate.symbol](atom, model, sigma)


def symbolic_matrix_multiplication(lhs: Matrix, rhs: Matrix):
    A, B = lhs.shape
    C, D = rhs.shape

    if B != C:
        raise TypeError('matrices {}x{} and {}x{} cannot be multiplied together'.format(A, B, C, D))

    zip_b = zip(*rhs.matrix)
    zip_b = list(zip_b)
    return [[sum(ele_a * ele_b for ele_a, ele_b in zip(row_a, col_b))
             for col_b in zip_b] for row_a in lhs.matrix]


def evaluate_builtin_function(term, model, sigma):
    bif = builtins.BuiltinFunctionSymbol
    ae1 = _arithmetic_evaluator_1
    ae2 = _arithmetic_evaluator_2
    _evaluators = {
        bif.ADD: lambda f, m, s: ae2(operator.add, f.subterms[0], f.subterms[1], m, s),
        bif.SUB: lambda f, m, s: ae2(operator.sub, f.subterms[0], f.subterms[1], m, s),
        bif.MUL: lambda f, m, s: ae2(operator.mul, f.subterms[0], f.subterms[1], m, s),
        bif.MATMUL: lambda f, m, s: ae2(symbolic_matrix_multiplication, f.subterms[0], f.subterms[1], m, s),
        bif.DIV: lambda f, m, s: ae2(operator.truediv, f.subterms[0], f.subterms[1], m, s),
        bif.POW: lambda f, m, s: ae2(operator.pow, f.subterms[0], f.subterms[1], m, s),
        bif.MOD: lambda f, m, s: ae2(operator.mod, f.subterms[0], f.subterms[1], model, s),
        bif.MIN: lambda f, m, s: ae2(funcsym.impl[bif.MIN.value], f.subterms[0], f.subterms[1], m, s),
        bif.MAX: lambda f, m, s: ae2(funcsym.impl[bif.MAX.value], f.subterms[0], f.subterms[1], m, s),
        bif.ABS: lambda f, m, s: ae1(funcsym.impl[bif.ABS.value], f.subterms[0], m, s),
        bif.SIN: lambda f, m, s: ae1(funcsym.impl[bif.SIN.value], f.subterms[0], m, s),
        bif.COS: lambda f, m, s: ae1(funcsym.impl[bif.COS.value], f.subterms[0], m, s),
        bif.TAN: lambda f, m, s: ae1(funcsym.impl[bif.TAN.value], f.subterms[0], m, s),
        bif.ATAN: lambda f, m, s: ae1(funcsym.impl[bif.ATAN.value], f.subterms[0], m, s),
        bif.ASIN: lambda f, m, s: ae1(funcsym.impl[bif.ASIN.value], f.subterms[0], m, s),
        bif.EXP: lambda f, m, s: ae1(funcsym.impl[bif.EXP.value], f.subterms[0], m, s),
        bif.LOG: lambda f, m, s: ae1(funcsym.impl[bif.LOG.value], f.subterms[0], m, s),
        bif.ERF: lambda f, m, s: ae1(funcsym.impl[bif.ERF.value], f.subterms[0], m, s),
        bif.ERFC: lambda f, m, s: ae1(funcsym.impl[bif.ERFC.value], f.subterms[0], m, s),
        bif.SGN: lambda f, m, s: ae1(funcsym.impl[bif.SGN.value], f.subterms[0], m, s),
        bif.SQRT: lambda f, m, s: ae1(funcsym.impl[bif.SQRT.value], f.subterms[0], m, s),
        bif.NORMAL: lambda f, m, s: ae2(funcsym.impl[bif.NORMAL.value], f.subterms[0], f.subterms[1], m, s),
        bif.GAMMA: lambda f, m, s: ae2(funcsym.impl[bif.GAMMA.value], f.subterms[0], f.subterms[1], m, s),
    }

    return _evaluators[term.symbol.symbol](term, model, sigma)


def _arithmetic_evaluator_1(operation, expr, model, sigma):
    # _lhs = args[0].symbol
    # _rhs = args[1].symbol
    # assert self.domain[0].contains(_lhs)
    # assert self.domain[1].contains(_rhs)
    expr = evaluate_term(expr, model, sigma)
    value = operation(ops.cast_to_number(expr))
    sort = ops.infer_numeric_sort(value, expr.language)
    return Constant(value, sort)


def _arithmetic_evaluator_2(operation, lhs, rhs, model, sigma):
    # _lhs = args[0].symbol
    # _rhs = args[1].symbol
    # assert self.domain[0].contains(_lhs)
    # assert self.domain[1].contains(_rhs)
    lhs = evaluate_term(lhs, model, sigma)
    rhs = evaluate_term(rhs, model, sigma)
    if isinstance(lhs, Matrix) and isinstance(rhs, Matrix):
        # print("Matrix op Matrix")
        value = operation(lhs, rhs)
        # print('Result : {}'.format(value))
        return evaluate_term(Matrix(value, lhs.sort), model, sigma)
    elif isinstance(lhs, Matrix) and not isinstance(rhs, Matrix):
        value = operation(lhs.matrix, ops.cast_to_number(rhs))
        return evaluate_term(Matrix(value, lhs.sort), model, sigma)
    elif isinstance(rhs, Matrix) and not isinstance(lhs, Matrix):
        value = operation(ops.cast_to_number(lhs), rhs.matrix)
        return evaluate_term(Matrix(value, rhs.sort), model, sigma)

    value = operation(ops.cast_to_number(lhs), ops.cast_to_number(rhs))
    sort = ops.infer_numeric_sort(value, lhs.language)
    return Constant(value, sort)



import numpy as np

from ...syntax import Term, Constant
from ...syntax.sorts import Sort
from ... import errors as err


class Matrix(Term):
    def __init__(self, arraylike, sort: Sort):
        self.matrix = np.array(arraylike, dtype=np.dtype(object))
        self._sort = sort
        # verify and cast
        N, M = self.matrix.shape
        for i in range(N):
            for j in range(M):
                m_ij = self.matrix[i, j]
                if isinstance(m_ij, Term):
                    if m_ij.sort != sort:
                        raise err.SyntacticError("Matrix: all \
                        entries need to be of sort '{}', \
                        entry ({},{}) is '{}'".format(sort.name, i, j, m_ij.sort.name))
                else:
                    self.matrix[i, j] = Constant(sort.cast(m_ij), sort)

    @property
    def shape(self):
        return self.matrix.shape

    @property
    def language(self):
        return self.sort.language

    @property
    def sort(self):
        return self._sort

    def __setitem__(self, key, value):
        self.matrix[key] = value

    def __getitem__(self, arg):
        i, j = arg
        return self.matrix[i, j]

    def __str__(self):
        return '{}'.format(self.matrix)

    __repr__ = __str__

    def hash(self):
        N, M = self.matrix.shape
        syms = []
        for i in range(N):
            for j in range(M):
                syms += [self.matrix[i, j]]
        return hash(tuple(syms) + self.matrix.shape)

    def is_syntactically_equal(self, other):
        if self.__class__ is not other.__class__ or self.matrix.shape != other.matrix.shape:
            return False
        N, M = self.matrix.shape
        for i in range(N):
            for j in range(M):
                if not self.matrix[i, j].is_syntactically_equal(other.matrix[i, j]):
                    return False
        return True




from setuptools import setup, find_packages
from codecs import open
from os import path
import importlib

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


# Load the version from ./src/tarski/version.py
spec = importlib.util.spec_from_file_location('tsk.version', path.join(here, 'src/tarski/version.py'))
version = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version)


def main():

    setup(
        name='tarski',
        version=version.__version__,
        description='Tarski is a framework for the specification, modeling and manipulation of AI planning problems.',
        long_description=long_description,
        long_description_content_type='text/markdown',

        url='https://github.com/aig-upf/tarski',
        author='Miquel Ramírez and Guillem Francès',
        author_email='guillem.frances@unibas.ch',

        keywords='planning logic STRIPS RDDL',
        classifiers=[
            'Development Status :: 3 - Alpha',

            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',

            'Topic :: Scientific/Engineering :: Artificial Intelligence',

            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
        ],

        packages=find_packages('src'),  # include all packages under src
        package_dir={'': 'src'},  # tell distutils packages are under src


        install_requires=[
            'antlr4-python3-runtime>=4.7.2',
            'numpy',
            'scipy',
            'pyrddl'
        ],

        extras_require={
            'dev': ['pytest', 'tox'],
            'test': ['pytest', 'tox'],
        },

        # This will include non-code files specified in the manifest, see e.g.
        # http://python-packaging.readthedocs.io/en/latest/non-code-files.html
        include_package_data=True
    )


if __name__ == '__main__':
    main()




import itertools

from ...fstrips import hybrid
from ...syntax import create_substitution, TermSubstitution
from ...util import IndexDictionary
from . import instantiation
from .elements import process_expression


class DifferentialConstraintGrounder:

    def __init__(self, prob, index):
        self.problem = prob
        self.L = self.problem.language
        self.index = index
        self.problem.ground_differential_constraints = IndexDictionary()
        self.schemas = list(self.problem.differential_constraints.values())
        self.differential_constraints_generated = 0

    def __str__(self):
        return 'Reactions generated: {}'.format(self.differential_constraints_generated)

    def calculate_constraints(self):

        for ode_schema in self.schemas:
            k, syms, substs = instantiation.enumerate_groundings(self.L, ode_schema.parameters)
            for values in itertools.product(*substs):
                subst = create_substitution(syms, values)
                op = TermSubstitution(self.L, subst)
                g_cond = process_expression(self.L, ode_schema.condition, op)
                g_variate = process_expression(self.L, ode_schema.variate, op)
                g_ode = process_expression(self.L, ode_schema.ode, op)

                self.problem.ground_differential_constraints.add(
                    hybrid.DifferentialConstraint(self.L, ode_schema.name, [], g_cond, g_variate, g_ode))
            self.differential_constraints_generated += k



"""
    Negation Builtin Rewriter
"""
import copy

from ..formulas import Connective, Atom, QuantifiedFormula, CompoundFormula
from ..builtins import create_atom


class NegatedBuiltinAbsorption:
    """
        This class rewrites the input formula phi into an equivalent formula
        absorbing the negation if the negated formula is a built-in
    """

    def __init__(self, lang, phi, do_copy=True):
        self.lang = lang
        self.blueprint = None
        if do_copy:
            self.blueprint = copy.deepcopy(phi)
        else:
            self.blueprint = phi
        self.formula = None

    def _convert(self, phi):
        if isinstance(phi, CompoundFormula):
            if phi.connective == Connective.Not:
                p = phi.subformulas[0]
                if isinstance(p, Atom):
                    if p.predicate.builtin:
                        try:
                            c = p.predicate.symbol.complement()
                            return create_atom(self.lang, c, p.subterms[0], p.subterms[1])
                        except (AttributeError, KeyError):
                            pass
                return phi

            else:
                assert phi.connective == Connective.And or phi.connective == Connective.Or
                new_sub = [self._convert(phi.subformulas[0]),
                           self._convert(phi.subformulas[1])]
                phi.subformulas = tuple(new_sub)
                return phi
        elif isinstance(phi, QuantifiedFormula):
            phi.formula = self._convert(phi.formula)
            return phi
        else:
            return phi

    def convert(self):
        self.formula = self._convert(self.blueprint)

    @staticmethod
    def rewrite(lang, phi, do_copy=True):
        trans = NegatedBuiltinAbsorption(lang, phi, do_copy)
        trans.convert()
        return trans



from ..builtins import BuiltinFunctionSymbol as BFS


def min(x, y):
    try:
        min_func = x.language.get_function(BFS.MIN)
    except AttributeError:
        try:
            min_func = y.language.get_function(BFS.MIN)
        except AttributeError:
            return min(x, y)
    return min_func(x, y)


def max(x, y):
    try:
        max_func = x.language.get_function(BFS.MAX)
    except AttributeError:
        try:
            max_func = y.language.get_function(BFS.MAX)
        except AttributeError:
            return max(x, y)
    return max_func(x, y)


def abs(x):
    abs_func = x.language.get_function(BFS.ABS)
    return abs_func(x)


def sin(x):
    sin_func = x.language.get_function(BFS.SIN)
    return sin_func(x)


def cos(x):
    cos_func = x.language.get_function(BFS.COS)
    return cos_func(x)


def tan(x):
    tan_func = x.language.get_function(BFS.TAN)
    return tan_func(x)


def atan(x):
    atan_func = x.language.get_function(BFS.ATAN)
    return atan_func(x)


def asin(x):
    asin_func = x.language.get_function(BFS.ASIN)
    return asin_func(x)


def exp(x):
    exp_func = x.language.get_function(BFS.EXP)
    return exp_func(x)


def log(x):
    log_func = x.language.get_function(BFS.LOG)
    return log_func(x)


def erf(x):
    erf_func = x.language.get_function(BFS.ERF)
    return erf_func(x)


def erfc(x):
    erfc_func = x.language.get_function(BFS.ERFC)
    return erfc_func(x)


def sgn(x):
    sgn_func = x.language.get_function(BFS.SGN)
    return sgn_func(x)



"""
 The loader imports the AST visitor from the parser directory - this directory is autogenerated by the
 utils/parser appropriate scripts, and should not be manually modified.
"""
import copy
import logging

from antlr4 import FileStream, CommonTokenStream, InputStream
from antlr4.error.ErrorListener import ErrorListener

from .common import parse_number, process_requirements, create_sort
from ...errors import SyntacticError
from ...fstrips import DelEffect, AddEffect, FunctionalEffect, UniversalEffect, OptimizationMetric
from ...syntax import CompoundFormula, Connective, neg, Tautology, implies, exists, forall, Term, Interval
from ...syntax.builtins import get_predicate_from_symbol, get_function_from_symbol
from ...syntax.formulas import VariableBinding

from .parser.visitor import fstripsVisitor
from .parser.lexer import fstripsLexer
from .parser.parser import fstripsParser


class FStripsParser(fstripsVisitor):
    """
    The parser assumes that the domain file is visited _before_ the instance file
    """

    def parse_string(self, string, start_rule='pddlDoc'):
        """ Parse a given string starting from a given grammar rule """
        return self._parse_stream(InputStream(string), start_rule)

    def parse_file(self, filename, start_rule='pddlDoc'):
        """ Parse a given filename starting from a given grammar rule """
        return self._parse_stream(FileStream(filename, encoding='utf-8'), start_rule)

    def _parse_stream(self, filestream, start_rule='pddlDoc'):
        lexer = self._configure_error_handling(fstripsLexer(filestream))
        stream = CommonTokenStream(lexer)
        parser = self._configure_error_handling(fstripsParser(stream))

        assert hasattr(parser, start_rule)
        tree = getattr(parser, start_rule)()
        return tree, stream

    def _configure_error_handling(self, element):
        if self.error_handler is not None:
            # If necessary, _replace_ previous error handlers with the given one
            element.removeErrorListeners()
            element.addErrorListener(self.error_handler)
        return element

    def __init__(self, problem, raise_on_error=False):
        self.error_handler = ExceptionRaiserListener() if raise_on_error else None
        self.current_binding = None

        # Shortcuts
        self.problem = problem
        self.language = problem.language
        self.init = problem.init

        self.requirements = set()

    def visitDomainName(self, ctx):
        self.problem.domain_name = ctx.NAME().getText().lower()

    def visitProblemDecl(self, ctx):
        self.problem.name = ctx.NAME().getText().lower()

    def visitProblemDomain(self, ctx):
        domain_name_as_declared_in_instance = ctx.NAME().getText().lower()
        if domain_name_as_declared_in_instance != self.problem.domain_name:
            logging.warning('Domain names as declared in domain and instance files do not coincide: "{}" vs " {}"'.
                            format(self.problem.domain_name, domain_name_as_declared_in_instance))

    def visitRequireDef(self, ctx):
        for req_ctx in ctx.REQUIRE_KEY():
            requirement = req_ctx.getText().lower()
            self.requirements.add(requirement)

        process_requirements(self.requirements, self.language)

    def visitDeclaration_of_types(self, ctx):
        for typename, basename in self.visit(ctx.possibly_typed_type_list()):
            create_sort(self.language, typename, basename)

    def extract_namelist(self, ctx):
        return [name.getText().lower() for name in ctx.NAME()]

    def visitSimpleNameList(self, ctx):
        names = self.extract_namelist(ctx)
        return [(name, 'object') for name in names]

    def visitName_list_with_type(self, ctx):
        typename = self.visit(ctx.typename())
        names = self.extract_namelist(ctx)
        return [(name, typename) for name in names]

    def visitPrimitiveTypename(self, ctx):
        return ctx.primitive_type().getText().lower()

    def visitEitherTypename(self, ctx):
        raise UnsupportedLanguageFeature('"either"-based types not supported in Tarski PDDL parser')

    def visitComplexNameList(self, ctx):
        untyped = self.visitSimpleNameList(ctx)
        typed = []
        for sub in ctx.name_list_with_type():
            typed += self.visit(sub)
        return untyped + typed

    def visitUntypedTypenameList(self, ctx):
        names = self.extract_namelist(ctx)
        return [(name, 'object') for name in names]

    def visitTypedTypenameList(self, ctx):
        untyped = self.visitUntypedTypenameList(ctx)
        typed = []
        for sub in ctx.name_list_with_type():
            typed += self.visit(sub)
        return untyped + typed

    def visitSingle_predicate_definition(self, ctx):
        predicate = ctx.predicate().getText().lower()
        argument_types = [a.sort for a in self.visit(ctx.possibly_typed_variable_list())]
        return self.language.predicate(predicate, *argument_types)

    def visitUntypedVariableList(self, ctx):
        variables = [self.language.variable(name.getText().lower(), 'object') for name in ctx.VARIABLE()]
        return variables

    def visitTypedVariableList(self, ctx):
        untyped_var_names = [self.language.variable(name.getText().lower(), 'object') for name in ctx.VARIABLE()]
        typed_var_names = []
        for sub_ctx in ctx.variable_list_with_type():
            typed_var_names += self.visit(sub_ctx)
        return typed_var_names + untyped_var_names

    def visitVariable_list_with_type(self, ctx):
        typename = self.visit(ctx.typename())  # This is the type of all variables in the list
        return [self.language.variable(name.getText().lower(), typename) for name in ctx.VARIABLE()]

    def visitTyped_function_definition(self, ctx, return_type=None):
        return_type = return_type or ctx.primitive_type().getText().lower()
        name = ctx.logical_symbol_name().getText().lower()
        argument_types = [a.sort for a in self.visit(ctx.possibly_typed_variable_list())]
        return self.language.function(name, *argument_types, return_type)

    def visitUntyped_function_definition(self, ctx):
        # According to Daniel Kovacs' PDDL 3.1 BNF spec, the default type for functions is 'number'
        return self.visitTyped_function_definition(ctx, 'number')

    def visitTypeBoundsDefinition(self, ctx):
        typename = ctx.NAME().getText().lower()
        sort = self.language.get_sort(typename)
        if not isinstance(sort, Interval):
            raise ParsingError("Attempt at bounding symbolic non-interval sort '{}'".format(sort))

        # Encode the bounds and set them into the sort
        lower = sort.encode(ctx.NUMBER(0).getText())
        upper = sort.encode(ctx.NUMBER(1).getText())
        sort.set_bounds(lower, upper)

    def visitObject_declaration(self, ctx):
        for o, t in self.visit(ctx.possibly_typed_name_list()):
            # TODO We might want to record elsewhere that these constants are
            # TODO required as per the PDDL spec to have fixed denotation
            self.language.constant(o, t)

    # For a fixed problem, there's no particular distinction btw domain constants and problem objects.
    visitConstant_declaration = visitObject_declaration

    def push_variables(self, variables, root=False):
        return ParserVariableContext(self, variables, root)

    def visitActionDef(self, ctx):
        name = ctx.actionName().getText().lower()
        params = self.visit(ctx.possibly_typed_variable_list())
        binding = VariableBinding(params)

        with self.push_variables(params, root=True) as _:
            precondition, effect = self.visit(ctx.actionDefBody())

        self.problem.action(name, binding, precondition, effect)

    def visitActionDefBody(self, ctx):
        prec = self.visit(ctx.precondition())
        eff = self.visit(ctx.effect())
        return prec, eff

    def visitTrivialPrecondition(self, ctx):
        return Tautology()

    def visitRegularPrecondition(self, ctx):
        return self.visit(ctx.goalDesc())

    def visitAtomicTermFormula(self, ctx):
        predicate_symbol = ctx.predicate().getText().lower()
        predicate = self.language.get_predicate(predicate_symbol)
        subterms = [self.visit(term_ctx) for term_ctx in ctx.term()]
        return predicate(*subterms)

    def visitTermGoalDesc(self, ctx):
        return self.visit(ctx.atomicTermFormula())

    def visitTermObject(self, ctx):
        name = ctx.NAME().getText().lower()
        return self.language.get_constant(name)

    def visitTermNumber(self, ctx):
        number = ctx.NUMBER().getText().lower()
        return parse_number(number, self.language)

    def _recover_variable_from_context(self, name):
        if self.current_binding is None:
            raise ParsingError("Variable '{}' used declared outside variable binding".format(name))

        return self.current_binding.get(name)

    def visitTermVariable(self, ctx):
        variable_name = ctx.VARIABLE().getText().lower()
        return self._recover_variable_from_context(variable_name)

    def visitGenericFunctionTerm(self, ctx):
        symbol = ctx.logical_symbol_name().getText().lower()
        func = self.language.get_function(symbol)
        subterms = [self.visit(t) for t in ctx.term()]
        return func(*subterms)

    def visitBinaryArithmeticFunctionTerm(self, ctx):
        op = ctx.builtin_binary_function().getText().lower()
        subterms = [self.visit(t) for t in ctx.term()]
        lhs, rhs = subterms
        return self.language.dispatch_operator(get_function_from_symbol(op), Term, Term, lhs, rhs)
    #
    # def visitUnaryArithmeticFunctionTerm(self, ctx):
    #     func_name = ctx.builtin_unary_function().getText().lower()
    #     if func_name not in built_in_functional_symbols:
    #         raise SystemExit("Function {0} first seen used as a term in an atomic formula".format(func_name))
    #     if func_name == '-':
    #         return FunctionalTerm('*', [self.visit(ctx.term()), NumericConstant(-1)])
    #     return FunctionalTerm(func_name, [self.visit(ctx.term())])

    def visitAndGoalDesc(self, ctx):
        conjuncts = [self.visit(sub_ctx) for sub_ctx in ctx.goalDesc()]
        # The PDDL spec allows for and AND with zero or a single conjunct (e.g. (and p), which Tarski does (rightly) not
        # We thus treat those cases specially.
        if len(conjuncts) == 0:
            return Tautology
        elif len(conjuncts) == 1:
            return conjuncts[0]
        return CompoundFormula(Connective.And, conjuncts)

    def visitOrGoalDesc(self, ctx):
        conjuncts = [self.visit(sub_ctx) for sub_ctx in ctx.goalDesc()]
        return CompoundFormula(Connective.Or, conjuncts)

    def visitNotGoalDesc(self, ctx):
        return neg(self.visit(ctx.goalDesc()))

    def visitImplyGoalDesc(self, ctx):
        lhs = self.visit(ctx.goalDesc(0))
        rhs = self.visit(ctx.goalDesc(1))
        return implies(lhs, rhs)

    def _visit_quantified_formula(self, ctx):
        variables = self.visit(ctx.possibly_typed_variable_list())

        with self.push_variables(variables, root=False) as _:
            formula = self.visit(ctx.goalDesc())
        return variables, formula

    def _visit_quantified_effect(self, ctx):
        """ Universally-quantified effects are different to formulas in that they are shorthand
            for a list of effects """
        variables = self.visit(ctx.possibly_typed_variable_list())

        with self.push_variables(variables, root=False) as _:
            formula = self.visit(ctx.effect())
        return variables, formula

    def visitExistentialGoalDesc(self, ctx):
        variables, formula = self._visit_quantified_formula(ctx)
        return exists(*variables, formula)

    def visitUniversalGoalDesc(self, ctx):
        variables, formula = self._visit_quantified_formula(ctx)
        return forall(*variables, formula)

    def visitBuiltinBinaryAtom(self, ctx):
        op = ctx.builtin_binary_predicate().getText().lower()
        lhs = self.visit(ctx.term(0))
        rhs = self.visit(ctx.term(1))
        return self.language.dispatch_operator(get_predicate_from_symbol(op), Term, Term, lhs, rhs)

    def visitGoal(self, ctx):
        self.problem.goal = self.visit(ctx.goalDesc())

    def visitSingleEffect(self, ctx):
        # The effect might already be a list if it derives from a multiple conditional effect.
        # Otherwise, we turn it into a list
        effect = self.visit(ctx.single_effect())
        return effect if isinstance(effect, list) else [effect]

    def visitConjunctiveEffectFormula(self, ctx):
        return [self.visit(sub_ctx) for sub_ctx in ctx.single_effect()]

    def visitAtomicEffect(self, ctx):
        return self.visit(ctx.atomic_effect())

    def visitAddAtomEffect(self, ctx):
        return AddEffect(self.visit(ctx.atomicTermFormula()))

    def visitDeleteAtomEffect(self, ctx):
        return DelEffect(self.visit(ctx.atomicTermFormula()))

    def visitAssignConstant(self, ctx):
        return FunctionalEffect(self.visit(ctx.functionTerm()), self.visit(ctx.term()))

    def visitUniversallyQuantifiedEffect(self, ctx):
        variables, effect = self._visit_quantified_effect(ctx)
        return UniversalEffect(variables, effect)

    def visitSingleConditionalEffect(self, ctx):
        effect = self.visit(ctx.atomic_effect())
        effect.condition = self.visit(ctx.goalDesc())
        return effect

    def visitMultipleConditionalEffect(self, ctx):
        condition = self.visit(ctx.goalDesc())
        effects = [self.visit(sub_ctx) for sub_ctx in ctx.atomic_effect()]
        for eff in effects:
            eff.condition = condition  # We simply copy the condition in each effect

        return effects

    def visitInit(self, ctx):
        # i.e. simply visit all node children
        for element_ctx in ctx.init_element():
            self.visit(element_ctx)

    def visitInitPositiveLiteral(self, ctx):
        predicate, subterms = self.visit(ctx.flat_atom())
        self.init.add(predicate, *subterms)

    def visitInitNegativeLiteral(self, ctx):
        # No need to do anything here, as atoms are assumed by default to be false
        pass

    def visitInitFunctionAssignment(self, ctx):
        fun, subterms = self.visit(ctx.flat_term())
        value = self.visit(ctx.constant_name())
        self.init.setx(fun(*subterms), value)

    def visitFlat_atom(self, ctx):
        predicate = self.language.get_predicate(ctx.predicate().getText().lower())
        subterms = tuple(self.visit(term_ctx) for term_ctx in ctx.constant_name())
        return predicate, subterms

    def visitFlat_term(self, ctx):
        func_name = self.language.get_function(ctx.function_name().getText().lower())
        subterms = tuple(self.visit(term_ctx) for term_ctx in ctx.constant_name())
        return func_name, subterms

    def visitSymbolic_constant(self, ctx):
        return self.language.get_constant(ctx.NAME().getText().lower())

    def visitNumeric_constant(self, ctx):
        number = ctx.NUMBER().getText().lower()
        return number

    # def visitExtensionalConstraintGD(self, ctx):
    #     arg_list = []
    #     for fn_ctx in ctx.groundFunctionTerm():
    #         arg_list.append(self.visit(fn_ctx))
    #     return [Atom(ctx.EXTNAME().getText().lower(), arg_list)]

    def visitAlternativeAlwaysConstraint(self, ctx):
        return [self.visit(ctx.goalDesc())]

    def visitConjunctionOfConstraints(self, ctx):
        constraints = []
        for conGD_ctx in ctx.prefConGD():
            constraints += self.visit(conGD_ctx)
        return constraints

    def visitPlainConstraintList(self, ctx):
        constraints = []
        for conGD_ctx in ctx.conGD():
            constraints += self.visit(conGD_ctx)
        return constraints

    # def visitProbConstraints(self, ctx):
    #     self.constraints = Conjunction(self.visit(ctx.prefConGD()))

    def visitProblemMetric(self, ctx):
        opt_type = ctx.optimization().getText().lower()
        opt_expression = self.visit(ctx.metricFExp())
        self.problem.metric = OptimizationMetric(opt_expression, opt_type)

    def visitFunctionalExprMetric(self, ctx):
        return None, self.visit(ctx.functionTerm())

    def visitCompositeMetric(self, ctx):
        return self.visit(ctx.terminalCost()), self.visit(ctx.stageCost())

    def visitTerminalCost(self, ctx):
        return self.visit(ctx.functionTerm())

    def visitStageCost(self, ctx):
        return self.visit(ctx.functionTerm())

    def visitTotalTimeMetric(self, ctx):
        raise UnsupportedLanguageFeature("Unsupported feature: Minimize total-time metric is not supported")

    def visitIsViolatedMetric(self, ctx):
        raise UnsupportedLanguageFeature("Unsupported feature: Count of violated constraints metric is not supported")

    def visitAssignEffect(self, ctx):
        operation = ctx.assignOp().getText().lower()
        lhs = self.visit(ctx.functionTerm())
        rhs = self.visit(ctx.term())
        operator = {'scale-up': '*', 'scale-down': '/', 'increase': '+', 'decrease': '-'}[operation]
        rhs = self.language.dispatch_operator(get_function_from_symbol(operator), Term, Term, lhs, rhs)
        return FunctionalEffect(lhs, self.visit(ctx.term()), rhs)

    def visitDerivedDef(self, ctx):
        raise UnsupportedLanguageFeature("Parsing of derived predicates in Tarski not yet implemented")


class UnresolvedVariableError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class UndeclaredVariable(Exception):
    def __init__(self, component, value):
        self.component = component
        self.value = value

    def __str__(self):
        return 'in {} found undeclared variable {}'.format(self.component, repr(self.value))


class ParserVariableContext:
    def __init__(self, parser: FStripsParser, pushed_variables, root=False):
        self.root = root
        self.parser = parser
        self.variables = pushed_variables
        self.previous_binding = None

    def __enter__(self):
        # Merge the new variable binding with the previous one, if it existed
        if self.root and self.parser.current_binding is not None:
            raise ParsingError("Clean ParserVariableContext opened upon existing context")

        if self.parser.current_binding is None:
            self.parser.current_binding = VariableBinding(self.variables)
        else:
            self.previous_binding = copy.deepcopy(self.parser.current_binding)
            for v in self.variables:
                self.parser.current_binding.add(v)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore the previous binding
        self.parser.current_binding = self.previous_binding


class ParsingError(SyntacticError):
    pass


class UnsupportedLanguageFeature(ParsingError):
    pass


class ExceptionRaiserListener(ErrorListener):
    """ An ANTLR ErrorListener that simply transforms any syntax error into a Tarski parsing error.
        Useful at least for testing purposes.
    """
    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        """ """
        msg = "line " + str(line) + ":" + str(column) + " " + msg
        raise ParsingError(msg)




import itertools


def get_symbols(lang, type_="all", include_builtin=True):
    """ Return all symbols of the given type (function/predicate), and including or excluding builtins as desired """
    if type_ == "predicate":
        iterate_over = lang.predicates
    elif type_ == "function":
        iterate_over = lang.functions
    else:
        assert type_ == "all"
        iterate_over = itertools.chain(lang.predicates, lang.functions)

    if include_builtin:
        return iterate_over
    else:
        return (x for x in iterate_over if not x.builtin)


def termlists_are_equal(terms1, terms2):
    """ Check whether two given lists of terms are (syntactic-wise) equal. """
    return len(terms1) == len(terms2) and all(x.is_syntactically_equal(y) for x, y in zip(terms1, terms2))


def termlist_hash(terms):
    """ Return a ready-to-hash tuple with the hashes of a list of terms. """
    return tuple(x.hash() for x in terms)



from . import errors as err


class Derived:
    """ A derived predicate. """

    def __init__(self, language, name, parameters, formula):
        # MRJ: note this will raise an exception if the predicate
        # has not been defined
        self.predicate = language.get_predicate(name)
        self.language = language
        # Check arity and type
        if len(parameters) != self.predicate.arity:
            raise err.InvalidDerivedPredicateError(name, formula,
                                                   "Arity of predicate does not match length of parameter list")
        for k, param in enumerate(parameters):
            if self.predicate.sort[k] != param.sort:
                raise err.InvalidDerivedPredicateError(name, formula, "Type mismatch of parameter \
            #{}, should be: {} provided: {}".format(k, self.predicate.sort.name, param.sort.name))
        self.parameters = parameters
        self.formula = formula

    def __lt__(self, other):
        return self.name < other.name

    def dump(self):
        return dict(name=self.predicate.symbol,
                    params=[par.dump() for par in self.parameters],
                    formula=self.formula.dump())

    def ident(self):
        params = ', '.join([str(o) for o in self.parameters])
        return '{}({})'.format(self.predicate.symbol, params)

    def __str__(self):
        tokens = ['derived {} {}:'.format(self.predicate.symbol,
                                          ' '.join([str(o) for o in self.parameters])),
                  'formula=({})'.format(self.precondition)]
        return '\n'.join(tokens)



""" Management of the theories (e.g. equality, etc.) associated to the FO languages """
from enum import Enum

from .fol import FirstOrderLanguage
from .syntax import builtins, Term
from .syntax.factory import create_atom, create_arithmetic_term
from .syntax.ops import cast_to_closest_common_ancestor
from . import errors as err


class Theory(Enum):
    """ """
    EQUALITY = "equality"
    ARITHMETIC = "arithmetic"
    SPECIAL = "special"
    RANDOM = "random"

    def __str__(self):
        return self.value


def language(name='L', theories=None):
    """ Build a language with the given name and configure it with the given theories """
    theories = theories or []
    lang = FirstOrderLanguage(name)
    _ = [load_theory(lang, t) for t in theories]
    return lang


def load_theory(lang, theory):
    """ Load one of the buillt-in theories into the given language """
    if lang.language_components_frozen:
        raise err.LanguageError("Cannot load theories once language elements have been defined")
    loaders = {
        Theory.EQUALITY: load_equality_theory,
        Theory.ARITHMETIC: load_arithmetic_theory,
        Theory.SPECIAL: load_special_theory,
        Theory.RANDOM: load_random_theory,
    }
    loader = loaders.get(theory)
    if loader is None:
        raise err.UnknownTheory(theory)

    loader(lang)
    lang.theories.append(theory)


def load_equality_theory(lang):
    # TODO We should create type-specific versions of each predicate / function.
    object_t = lang.get_sort('object')
    for pred in builtins.get_equality_predicates():
        lang.register_operator_handler(pred, Term, Term, create_casting_handler(pred, create_atom))
        p = lang.predicate(pred, object_t, object_t)
        p.builtin = True


def load_arithmetic_theory(lang):
    # MRJ: ite function is now part of the Arithmetic theory
    # "virtual function" ite (if-then-else) registered
    ite_func = lang.function('ite', lang.Object, lang.Object, lang.Object)
    ite_func.builtin = True

    fun = builtins.BuiltinFunctionSymbol.MATMUL
    lang.register_operator_handler(fun, Term, Term, create_casting_handler(fun, create_arithmetic_term))
    f = lang.function(fun, lang.Real, lang.Real, lang.Real)
    f.builtin = True

    for pred in builtins.get_arithmetic_predicates():
        lang.register_operator_handler(pred, Term, Term, create_casting_handler(pred, create_atom))
        p = lang.predicate(pred, lang.Real, lang.Real)
        p.builtin = True

    for fun in builtins.get_arithmetic_binary_functions():
        lang.register_operator_handler(fun, Term, Term, create_casting_handler(fun, create_arithmetic_term))
        f = lang.function(fun, lang.Real, lang.Real, lang.Real)
        f.builtin = True

    for fun in builtins.get_arithmetic_unary_functions():
        lang.register_unary_operator_handler(fun, Term, create_casting_handler(fun, create_arithmetic_term))
        f = lang.function(fun, lang.Real, lang.Real)
        f.builtin = True


def load_special_theory(lang):
    for fun in builtins.get_special_binary_functions():
        lang.register_operator_handler(fun, Term, Term, create_casting_handler(fun, create_arithmetic_term))
        f = lang.function(fun, lang.Real, lang.Real, lang.Real)
        f.builtin = True
    for fun in builtins.get_special_unary_functions():
        lang.register_unary_operator_handler(fun, Term, create_casting_handler(fun, create_arithmetic_term))
        f = lang.function(fun, lang.Real, lang.Real)
        f.builtin = True


def load_random_theory(lang):
    for fun in builtins.get_random_binary_functions():
        lang.register_operator_handler(fun, Term, Term, create_casting_handler(fun, create_arithmetic_term))
        f = lang.function(fun, lang.Real, lang.Real, lang.Real)
        f.builtin = True
    for fun in builtins.get_random_unary_functions():
        lang.register_unary_operator_handler(fun, Term, create_casting_handler(fun, create_arithmetic_term))
        f = lang.function(fun, lang.Real, lang.Real)
        f.builtin = True


def create_casting_handler(symbol, factory_method):
    """ """
    def handler(lhs, rhs):
        lhs, rhs = cast_to_closest_common_ancestor(lhs, rhs)
        return factory_method(symbol, lhs, rhs)
    return handler




from ..syntax import symref, CompoundFormula, QuantifiedFormula, Atom, CompoundTerm, Variable


class MatchExpression:
    """
        This Visitor traverses a formula/expression stopping when
        finding an exact syntactic match for the target
    """

    def __init__(self, target):
        self.psi = target
        self.target = symref(target)
        self.hits = 0

    def _check_match(self, phi):
        if symref(phi) == self.target:
            self.hits += 1
            return True
        return False

    def visit(self, phi):
        if isinstance(phi, CompoundFormula):
            if self._check_match(phi):
                return
            _ = [self.visit(f) for f in phi.subformulas]

        elif isinstance(phi, QuantifiedFormula):
            if self._check_match(phi):
                return
            self.visit(phi.formula)

        elif isinstance(phi, (Atom, CompoundTerm)):
            if self._check_match(phi):
                return
            _ = [self.visit(f) for f in phi.subterms]


class CollectVariables:
    """ Collect all variables in a given formula or term """

    def __init__(self):
        self.variables = set()

    def visit(self, phi):
        if isinstance(phi, Variable):
            self.variables.add(phi)

        elif isinstance(phi, CompoundFormula):
            _ = [self.visit(f) for f in phi.subformulas]

        elif isinstance(phi, QuantifiedFormula):
            self.visit(phi.formula)

        elif isinstance(phi, (Atom, CompoundTerm)):
            _ = [self.visit(f) for f in phi.subterms]


class CollectFreeVariables:
    """ Collect all free variables in a given formula or term """
    def __init__(self):
        self.quantified_vars = set()
        self._free_variables = set()

    @property
    def free_variables(self):
        for ref in self._free_variables:
            yield ref.expr

    def visit(self, phi):
        if isinstance(phi, Variable):
            t_ref = symref(phi)
            if t_ref not in self.quantified_vars:
                self._free_variables.add(t_ref)
        elif isinstance(phi, CompoundFormula):
            _ = [self.visit(f) for f in phi.subformulas]

        elif isinstance(phi, QuantifiedFormula):
            _ = [self.quantified_vars.add(symref(x)) for x in phi.variables]
            self.visit(phi.formula)
            _ = [self.quantified_vars.remove(symref(x)) for x in phi.variables]

        elif isinstance(phi, (Atom, CompoundTerm)):
            _ = [self.visit(f) for f in phi.subterms]



from enum import Enum

# MRJ: Table with negated counterparts of builtin predicates. Used
# by the method ```complement``` of enum class BuiltinPredicateSymbol.
symbol_complements = {"=": "!=", "!=": "=", "<": ">=", "<=": ">", ">": "<=", ">=": "<"}


class BuiltinPredicateSymbol(Enum):
    EQ = "="
    NE = "!="
    LT = "<"
    LE = "<="
    GT = ">"
    GE = ">="

    def __str__(self):
        return self.value.lower()

    def complement(self):
        return BuiltinPredicateSymbol(symbol_complements[self.value])


class BuiltinFunctionSymbol(Enum):
    ADD = "+"
    SUB = "-"
    MUL = "*"
    MATMUL = "@"
    DIV = "/"
    POW = "**"
    MOD = "%"
    MIN = "min"
    MAX = "max"
    ABS = "abs"
    SIN = "sin"
    COS = "cos"
    TAN = "tan"
    ATAN = "atan"
    ASIN = "asin"
    EXP = "exp"
    LOG = "log"
    ERF = "erf"
    ERFC = "erfc"
    SGN = "sgn"
    SQRT = "sqrt"
    NORMAL = "normal"
    GAMMA = "gamma"
    KRON = "kron"
    DIRAC = "dirac"
    BERNOULLI = "bernoulli"
    DISCRETE = "discrete"
    POISSON = "poisson"

    # ...

    def __str__(self):
        return self.value.lower()


def is_builtin_predicate(predicate):
    return isinstance(predicate.symbol, BuiltinPredicateSymbol)


def create_atom(lang, symbol: BuiltinPredicateSymbol, lhs, rhs):
    from .formulas import Atom
    predicate = lang.get_predicate(symbol)
    return Atom(predicate, [lhs, rhs])


def is_builtin_function(fun):
    return isinstance(fun.symbol, BuiltinFunctionSymbol)

    # language = lhs.language
    # if language != rhs.language:
    #     raise err.LanguageMismatch(rhs, rhs.language, language)


def get_equality_predicates():
    return [BuiltinPredicateSymbol.EQ, BuiltinPredicateSymbol.NE]

    # TODO AT THE MOMENT WE DO NOT CHECK FOR TYPE SAFETY WITH BUILT-IN TYPES


def get_arithmetic_predicates():
    return [BuiltinPredicateSymbol.LT, BuiltinPredicateSymbol.LE, BuiltinPredicateSymbol.GT, BuiltinPredicateSymbol.GE]


def get_arithmetic_binary_functions():
    return [BuiltinFunctionSymbol.ADD, BuiltinFunctionSymbol.SUB, BuiltinFunctionSymbol.MUL, BuiltinFunctionSymbol.DIV,
            BuiltinFunctionSymbol.POW, BuiltinFunctionSymbol.MOD]


def get_arithmetic_unary_functions():
    return [BuiltinFunctionSymbol.SQRT]


def get_special_binary_functions():
    BFS = BuiltinFunctionSymbol
    return [BFS.MIN, BFS.MAX]


def get_special_unary_functions():
    BFS = BuiltinFunctionSymbol
    return [BFS.ABS, BFS.SIN, BFS.COS, BFS.TAN, BFS.ATAN, BFS.ASIN, BFS.EXP, BFS.LOG, BFS.ERF, BFS.ERFC, BFS.SGN]


def get_random_binary_functions():
    BFS = BuiltinFunctionSymbol
    return [BFS.NORMAL, BFS.GAMMA]


def get_random_unary_functions():
    # BFS = BuiltinFunctionSymbol
    return []


def get_predicate_from_symbol(symbol: str):
    return BuiltinPredicateSymbol(symbol)


def get_function_from_symbol(symbol: str):
    return BuiltinFunctionSymbol(symbol)

# def eq(lhs, rhs):
#     return create_atom(BuiltinPredicate.EQ, lhs, rhs)
#
#
# def ne(lhs, rhs):
#     return create_atom(BuiltinPredicate.NE, lhs, rhs)
#
#
# def lt(lhs, rhs):
#     return create_atom(BuiltinPredicate.LT, lhs, rhs)
#
#
# def gt(lhs, rhs):
#     return create_atom(BuiltinPredicate.GT, lhs, rhs)
#
#
# def le(lhs, rhs):
#     return create_atom(BuiltinPredicate.LE, lhs, rhs)
#
#
# def ge(lhs, rhs):
#     return create_atom(BuiltinPredicate.GE, lhs, rhs)




import logging

from .. import FirstOrderLanguage
from ..syntax import builtins
from . import Concept, Role, UniversalConcept, PrimitiveConcept, NotConcept, ExistsConcept, ForallConcept, \
    EqualConcept, PrimitiveRole, RestrictRole, AndConcept, EmptyConcept, CompositionRole, NominalConcept, NullaryAtom, \
    GoalNullaryAtom, GoalConcept, GoalRole, OrConcept


def filter_subnodes(elem, t):
    """ Return a list with all subnodes of a certain concept or role which conform to the given Python type """
    return list(filter(lambda x: type(x) == t, elem.flatten()))


def compute_dl_vocabulary(lang):
    """ Return the DL vocabulary for the given language.
    This is the list of all predicates of arity 0, 1 and 2, all types, and all functions of arity 0 and 1
    """
    v = [(p.symbol, p) for p in lang.predicates if not builtins.is_builtin_predicate(p) and 0 <= p.arity <= 2] +\
        [(f.symbol, f) for f in lang.functions if not builtins.is_builtin_function(f) and f.arity in (0, 1)] + \
        [(s.name, s) for s in lang.sorts if not s.builtin]
    return dict(v)


class SyntacticFactory:
    def __init__(self, language: FirstOrderLanguage):
        self.language = language
        self.universe_sort = language.get_sort('object')
        self.top = UniversalConcept(self.universe_sort.name)
        self.bot = EmptyConcept(self.universe_sort.name)

    def generate_primitives_from_language(self, nominals, types, goal_predicates):
        """ Generate primitive concepts from the language taking into account only the given nominals """

        concepts, roles, primitive_atoms = [], [], []
        for predfun in list(self.language.predicates) + list(self.language.functions):
            if builtins.is_builtin_predicate(predfun) or builtins.is_builtin_function(predfun):
                # Skip "=" and other built-in symbols
                # TODO We might want to revise this and allow for certain builtins
                continue

            if predfun.uniform_arity() == 0:
                primitive_atoms.append(NullaryAtom(predfun))
                if predfun.symbol in goal_predicates:
                    primitive_atoms.append(GoalNullaryAtom(predfun))

            elif predfun.uniform_arity() == 1:
                concepts.append(PrimitiveConcept(predfun))
                if predfun.symbol in goal_predicates:
                    concepts.append(GoalConcept(predfun))

            elif predfun.uniform_arity() == 2:
                roles.append(PrimitiveRole(predfun))
                if predfun.symbol in goal_predicates:
                    roles.append(GoalRole(predfun))

            else:
                logging.warning('Predicate/Function "{}" with normalized arity > 2 ignored'.format(predfun))

        for c in nominals:
            concepts.append(NominalConcept(c.symbol, c.sort))

        for t in types:
            concepts.append(PrimitiveConcept(t))

        logging.info('Primitive (nullary) atoms : {}'.format(", ".join(map(str, primitive_atoms))))
        logging.info('Primitive (unary) concepts: {}'.format(", ".join(map(str, concepts))))
        logging.info('Primitive (binary) roles  : {}'.format(", ".join(map(str, roles))))

        return concepts, roles, primitive_atoms

    def create_exists_concept(self, role: Role, concept: Concept):
        # exists(R.C) = { x | exists y R(x,y) and C(y) }

        result = ExistsConcept(role, concept)
        _, s2 = role.sort

        if concept == self.bot:
            logging.debug('Concept "{}" is statically empty'.format(result))
            return None

        # TODO ADD: If C is a sort-concept of the same sort than s2, then the concept will be equiv to exist(R.True)
        if not self.language.are_vertically_related(s2, concept.sort):
            logging.debug('Concept "{}" pruned for type-inconsistency reasons'.format(result))
            return None

        if isinstance(role, RestrictRole) and concept == self.top:
            return None  # Is syntactically equivalent to a simpler exists concept

        return result

    def create_not_concept(self, concept: Concept):
        if isinstance(concept, NotConcept):
            return None

        result = NotConcept(concept, self.universe_sort)
        return result

    def create_forall_concept(self, role: Role, concept: Concept):
        # forall(R.C) = { x | forall y R(x,y) implies C(y) }

        result = ForallConcept(role, concept)
        _, s2 = role.sort

        if isinstance(concept, UniversalConcept):
            # logging.debug('Concept "{}" equivalent to simpler concept'.format(result))
            return None

        if not self.language.are_vertically_related(s2, concept.sort):
            logging.debug('Concept "{}" pruned for type-inconsistency reasons'.format(result))
            return None

        return result

    def create_and_concept(self, c1: Concept, c2: Concept):
        # C1 AND C2 = { x | C1(x) AND C2(x) }

        sort = self.language.most_restricted_type(c1.sort, c2.sort)

        if c1 == c2:
            return None  # No sense in C and C

        if c1 in (self.top, self.bot) or c2 in (self.top, self.bot):
            logging.debug('AND of {} and {} pruned, no sense in AND\'ing with top or bot'.format(c1, c2))
            return None

        if sort is None:
            # i.e. c1 and c2 are disjoint types
            logging.debug('AND of {} and {} pruned for type-inconsistency reasons'.format(c1, c2))
            return None

        return AndConcept(c1, c2, sort)

    def create_or_concept(self, c1: Concept, c2: Concept):
        sort = self.language.most_restricted_type(c1.sort, c2.sort)

        if c1 == c2:
            return None  # No sense in C OR C

        if c1 in (self.top, self.bot) or c2 in (self.top, self.bot):
            logging.debug('OR of {} and {} pruned, no sense in OR\'ing with top or bot'.format(c1, c2))
            return None

        return OrConcept(c1, c2, sort)

    def create_equal_concept(self, r1: Role, r2: Role):
        assert isinstance(r1, Role) and isinstance(r2, Role)
        # The sort of the resulting concept will be the most restricted sort between the left sorts of the two roles
        sort = self.language.most_restricted_type(r1.sort[0], r2.sort[0])

        if sort is None:
            logging.debug('Concept "EqualConcept({},{})" pruned for type-inconsistency reasons'.format(r1, r2))
            return None
        return EqualConcept(r1, r2, sort)

    def create_restrict_role(self, r: Role, c: Concept):

        result = RestrictRole(r, c)
        if not self.language.are_vertically_related(r.sort[1], c.sort):
            logging.debug('Role "{}" pruned for type-inconsistency reasons'.format(result))
            return None

        if isinstance(c, UniversalConcept) or c == self.bot:
            logging.debug('Role "{}" pruned; no sense in restricting to top / bot concepts'.format(result))
            return None

        if isinstance(r, RestrictRole):
            logging.debug('Role "{}" pruned; no direct nesting of restrictions'.format(result))
            return None

        return result

    def create_composition_role(self, r1: Role, r2: Role):

        # Compose only on primitives or their inversions
        # if (not isinstance(r1, (PrimitiveRole, InverseRole)) or
        #    not isinstance(r2, (PrimitiveRole, InverseRole))):
        #     return None

        result = CompositionRole(r1, r2)

        if not self.language.are_vertically_related(r1.sort[1], r2.sort[0]):
            logging.debug('Role "{}" pruned for type-inconsistency reasons'.format(result))
            return None

        num_comp = len(filter_subnodes(result, CompositionRole))
        if num_comp > 2:
            logging.debug('Role "{}" pruned: number of compositions ({}) exceeds threshold'.format(result, num_comp))
            return None

        return result
